"""Test runfrog command line scripts and options."""

import sys
from pathlib import Path
from typing import Any

import pytest

from limax import EXAMPLE_LIMAX_PATIENT1_PATH, commands


@pytest.mark.parametrize("filename", [EXAMPLE_LIMAX_PATIENT1_PATH])
def test_limax_1(monkeypatch: Any, tmp_path: Path, filename: str) -> None:
    """Test command.

    limax --input src/resources/patient1.csv
    --output src/resources/limax_example_processed.csv
    """
    output_path = tmp_path / "processed.csv"
    with monkeypatch.context() as m:
        args = [
            "runfrog",
            "--input",
            filename,
            "--output",
            output_path,
        ]
        m.setattr(sys, "argv", args)
        commands.main()

        assert output_path.exists()



"""Rich console for logging."""

from rich import pretty
from rich.console import Console
from rich.theme import Theme


pretty.install()
custom_theme = Theme(
    {
        "success": "green",
        "info": "blue",
        "warning": "orange3",
        "error": "red",
    }
)

console = Console(record=True, theme=custom_theme)



"""limax - Python utilities for limax."""

from pathlib import Path

__author__ = "Matthias Koenig"
__version__ = "0.4.3"
__citation__ = "https://doi.org/10.5281/zenodo.7382669"


program_name: str = "limax"

RESOURCES_DIR: Path = Path(__file__).parent / "resources"
RAW_DIR: Path = RESOURCES_DIR / "raw"
PROCESSED_DIR: Path = RESOURCES_DIR / "processed"

EXAMPLE_LIMAX_PATIENT1_PATH: Path = RAW_DIR / "patient1.csv"
EXAMPLE_LIMAX_PATIENT2_PATH: Path = RAW_DIR / "patient2.csv"
EXAMPLE_LIMAX_PATIENT3_PATH: Path = RAW_DIR / "patient3.csv"
EXAMPLE_LIMAX_PATIENT2017_PATH: Path = RAW_DIR / "example_anonym_2017.csv"
EXAMPLE_LIMAX_PATIENT2018_PATH: Path = RAW_DIR / "example_anonym_2018.csv"
EXAMPLE_LIMAX_PATIENT2022_PATH: Path = RAW_DIR / "example_anonym_2022.csv"



"""Testing examples."""

from limax.example import example_limax_dir, example_limax_file


def test_example_limax_dir() -> None:
    """Run example 1."""
    example_limax_dir()


def test_example_limax_file() -> None:
    """Run example 2."""
    example_limax_file()



"""Test LiMAx IO functionality."""

import json
from pathlib import Path

from limax import EXAMPLE_LIMAX_PATIENT1_PATH
from limax.io import parse_limax_file
from limax.model import LX


def test_convert_limax_io(tmp_path: Path) -> None:
    """Test conversion functionality."""
    output_path: Path = tmp_path / "output.csv"
    lx = parse_limax_file(limax_csv=EXAMPLE_LIMAX_PATIENT1_PATH, output_dir=output_path)
    assert lx
    df = lx.data.to_df()
    assert not df.empty
    assert "time" in df.columns
    assert output_path.exists()


def test_json_serialization(tmp_path: Path) -> None:
    """Test roundtrip to JSON."""
    lx = parse_limax_file(limax_csv=EXAMPLE_LIMAX_PATIENT1_PATH, output_dir=tmp_path)
    assert lx
    output_path = tmp_path / f"{EXAMPLE_LIMAX_PATIENT1_PATH.stem}.json"
    assert output_path.exists()
    with open(output_path, "r") as f_json:
        json_data = json.load(f_json)
        lx2 = LX(**json_data)
        assert lx2



"""Models for LiMAx data."""

from typing import Any, Dict, List

import pandas as pd
from pydantic import BaseModel, Field


class LXMetaData(BaseModel):
    """LiMAx metadata.

    0 # mID 102
    1 # 'doc' (, )
    2 # Dr. Max Mustermann
    3 # 01.01.2010 08:30
    4 # utouARg
    5 # 160 cm
    6 # 70 kg
    7 # 43,295187
    8 # 44,395187
    9 # 630,0
    10 # Nahrungskarenz: über 3 Std., Raucher: Nein, Sauerstoff: Nein, Beatmung: Nein, Medikation: Ja
    """

    mid: str
    datetime: str = Field(repr=True)
    height: float = Field(description="Height in [cm]")
    weight: float = Field(description="Weight in [kg]")
    sex: str = Field(description="Sex in {M, F, NA}")
    smoking: bool = Field(description="Smoking status")
    oxygen: bool = Field(repr=True)
    ventilation: bool = Field(repr=True)
    medication: bool = Field(repr=True)
    food_abstinence: str = Field(repr=True)
    values: List[float] = Field(repr=True)
    comments: List[str] = Field(repr=False)


class LXData(BaseModel):
    """LiMAx data."""

    time: List[float]
    dob: List[float]
    error: List[str]

    def to_df(self) -> pd.DataFrame:
        """Get pandas DataFrame representation."""
        d: Dict[str, Any] = {
            "time": self.time,
            "dob": self.dob,
            "error": self.error,
        }
        df = pd.DataFrame(d)
        df = df[["time", "dob", "error"]]
        # make columns numeric
        # df = pd.to_numeric(df)
        # print(df.head())

        # sort by time (some strange artefacts in some files)
        df.sort_values(by=["time"], inplace=True)
        return df


class LX(BaseModel):
    """LiMAx DOB curve."""

    metadata: LXMetaData
    data: LXData = Field(repr=False)

    model_config: Dict = {"arbitrary_types_allowed": True}

    def to_df(self) -> pd.DataFrame:
        """Convert to DOB dataframe."""
        return self.data.to_df()



"""Definition of command line commands for limax."""

import argparse
from pathlib import Path

from limax import __citation__, __version__, log
from limax.console import console
from limax.io import parse_limax_file, read_limax_dir


logger = log.get_logger(__name__)


def main() -> None:
    """Entry point which runs LiMAx script.

    The script is registered as `limax` command.

    Example (process single raw file):
        limax -i src/limax/resources/patient1.csv -o src/limax/resources/limax_example_processed.csv

    Example (process all limax raw file in folder):
        limax -i src/limax/resources -o src/limax/resources
    """

    import optparse
    import sys

    parser = optparse.OptionParser()
    parser.add_option(
        "-i",
        "--input",
        action="store",
        dest="input_path",
        help="Path to folder with LiMAx raw files or single LiMAx raw file as '*.csv'.",
    )
    parser.add_option(
        "-o",
        "--output_dir",
        action="store",
        dest="output_dir",
        help="Path to output folder with processed LiMAx files as '*.json'.",
    )

    console.rule(style="white")
    console.print(":syringe: LIMAX ANALYSIS :syringe:")
    console.print(f"Version {__version__} (https://github.com/matthiaskoenig/limax)")
    console.print(f"Citation {__citation__}")
    console.rule(style="white")
    console.print("Example processing single LiMAx raw file:")
    console.print("    limax -i patient1.csv -o .")
    console.print("Example processing folder with LiMAx raw files:")
    console.print("    limax -i limax_examples -o limax_examples_processed")
    console.rule(style="white")

    options, args = parser.parse_args()

    def _parser_message(text: str) -> None:
        console.print(text)
        parser.print_help()
        console.rule(style="white")
        sys.exit(1)

    if not options.input_path:
        _parser_message("Required argument '--input' missing.")
    if not options.output_dir:
        _parser_message("Required argument '--output_dir' missing.")

    output_dir: Path = Path(options.output_dir)
    input_path: Path = Path(options.input_path)

    if not input_path:
        _parser_message(f"Input path does not exist: '--input {input_path}.")

    # process single LiMAx
    if input_path.is_file():
        parse_limax_file(limax_csv=input_path, output_dir=output_dir)

    # process folder with LiMAx raw data
    elif input_path.is_dir():
        # process all files
        read_limax_dir(input_dir=input_path, output_dir=output_dir)
    else:
        _parser_message(
            f"Input path is neither file nor directory: " f"'--input {input_path}."
        )


if __name__ == "__main__":
    main()



"""Visualization of DOB curves."""

from pathlib import Path
from typing import Optional

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from limax.model import LX


# ------------------------------------------------------------------------------
SMALL_SIZE = 12
MEDIUM_SIZE = 15
BIGGER_SIZE = 25

matplotlib.rc("font", size=SMALL_SIZE)  # controls default text sizes
matplotlib.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
matplotlib.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
# ------------------------------------------------------------------------------


def plot_lx_matplotlib(lx: LX, fig_path: Optional[Path] = None) -> None:
    """Plot DOB curve using matplotlib."""
    metadata = lx.metadata
    ax1: Axes
    ax2: Axes
    f: Figure
    f, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
    f.subplots_adjust(hspace=0.3)

    f.suptitle(
        f"mid: {lx.metadata.mid}",
        fontdict={
            "fontsize": 11,
        },
    )

    sex_text = "∅ (NA)"
    if metadata.sex == "M":
        sex_text = "♂ (M)"
    elif metadata.sex == "F":
        sex_text = "♀ (F)"
    text = "\n".join(
        [
            f"{'mid':<16}: {metadata.mid}",
            f"{'datetime':<16}: {metadata.datetime}",
            f"{'height':<16}: {metadata.height}",
            f"{'weight':<16}: {metadata.weight}",
            f"{'sex':<16}: {sex_text}",
            f"{'smoking':<16}: {'✓ (Yes)' if metadata.smoking else 'x (No)'}",
            f"{'oxygen':<16}: {'✓ (Yes)' if metadata.oxygen else 'x (No)'}",
            f"{'ventilation':<16}: {'✓ (Yes)' if metadata.ventilation else 'x (No)'}",
            f"{'medication':<16}: {'✓ (Yes)' if metadata.medication else 'x (No)'}",
            f"{'food abstinence':<16}: {metadata.food_abstinence}",
        ]
    )
    ax1.annotate(
        text,
        xy=(0.05, 0.4),
        xycoords="figure fraction",
        family="monospace",
        size=14,
    )

    for ax in (ax2, ax3):
        ax.plot(
            np.array(lx.data.time) / 60,
            lx.data.dob,
            "-o",
            color="black",
            markeredgecolor="black",
            markerfacecolor="tab:blue",
        )
        ax.grid(True)

        ax.set_xlabel("Time [min]", fontdict={"weight": "bold"})

    ax1.get_xaxis().set_visible(False)
    ax1.get_yaxis().set_visible(False)
    ax1.spines["top"].set_visible(False)
    ax1.spines["right"].set_visible(False)
    ax1.spines["bottom"].set_visible(False)
    ax1.spines["left"].set_visible(False)

    ax2.set_ylabel("DOB", fontdict={"weight": "bold"})
    ax3.set_yscale("log")
    ax3.set_ylim(bottom=1.0)

    plt.show()
    if fig_path:
        f.savefig(fig_path, dpi=300, bbox_inches="tight")


if __name__ == "__main__":
    from limax import EXAMPLE_LIMAX_PATIENT1_PATH
    from limax.io import parse_limax_file

    lx: LX = parse_limax_file(EXAMPLE_LIMAX_PATIENT1_PATH)
    print(lx.to_df())
    plot_lx_matplotlib(lx)



"""Examples for limax processing."""

from limax import EXAMPLE_LIMAX_PATIENT1_PATH, PROCESSED_DIR, RAW_DIR
from limax.io import parse_limax_file, read_limax_dir


def example_limax_file() -> None:
    """Run processing of single file."""
    parse_limax_file(EXAMPLE_LIMAX_PATIENT1_PATH, PROCESSED_DIR)


def example_limax_dir() -> None:
    """Run processing of folder."""
    read_limax_dir(RAW_DIR, PROCESSED_DIR)


if __name__ == "__main__":
    example_limax_file()
    example_limax_dir()



"""
Reading data from RAW Limax files.

Anonymization.
"""

import re
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from limax import log
from limax.console import console
from limax.model import LX, LXData, LXMetaData
from limax.plot import plot_lx_matplotlib


logger = log.get_logger(__file__)


def read_limax_dir(input_dir: Path, output_dir: Path) -> None:
    """Read limax data from folder."""
    # process all files
    lx_metadata = []
    for limax_csv in input_dir.glob("**/*.csv"):
        limax_csv_rel = limax_csv.relative_to(input_dir)
        output_path: Path = Path(output_dir / limax_csv_rel)
        lx: LX = parse_limax_file(limax_csv=limax_csv, output_dir=output_path.parent)
        lx_metadata.append(lx.metadata.model_dump())

    # create overall report
    df = pd.DataFrame(data=lx_metadata)
    df.to_csv(output_dir / "report.tsv", sep="\t", index=False)
    df_clean = df.drop(["comments"], axis=1)
    console.rule(style="[white]")
    console.log(df_clean)
    console.rule(style="[white]")


class MetadataParser:
    """Helper class to parse LiMAx metadata."""

    patient_metadata_fields = [
        "food_abstinence",
        "smoking",
        "oxygen",
        "ventilation",
        "medication",
    ]

    pattern_datetime = re.compile(r"\d{2}\.\d{2}\.\d{4}\s\d{2}:\d{2}")
    pattern_value = re.compile(r"^\d+,{0,1}\d*$")

    @classmethod
    def parse(cls, md_lines: List[str]) -> LXMetaData:
        """Parse metadata from metadata lines."""

        # find lines for information
        line_mid: Optional[str] = None
        line_datetime: Optional[str] = None
        line_pmd: Optional[str] = None
        line_weight: Optional[str] = None
        line_height: Optional[str] = None
        line_sex: Optional[str] = None
        lines_value: List[str] = []
        lines_comment: List[str] = []  # all lines after 'Nahrungskarenz'

        k_pmd: int = np.iinfo("int16").max
        for k, line in enumerate(md_lines):
            if k == 0:
                line_mid = line
            elif re.match(pattern=cls.pattern_datetime, string=line):
                # 01.01.2017 17:17
                line_datetime = line
            elif line.startswith("Nahrungskarenz"):
                line_pmd = line
                k_pmd = k
            elif line.endswith(" cm"):
                line_height = line
            elif line.endswith(" kg"):
                line_weight = line
            elif line == "männlich":
                line_sex = line
            elif line == "weiblich":
                line_sex = line
            elif re.match(pattern=cls.pattern_value, string=line):
                lines_value.append(line)
            elif k > k_pmd:
                # final comment lines
                lines_comment.append(line)
            elif k > 7:
                # additional lines which were not caught
                lines_comment.append(line)

        md_dict: Dict[str, Any] = {
            "mid": cls._parse_mid(line_mid),
            "datetime": cls._parse_datetime(line_datetime),
            "sex": cls._parse_sex(line_sex),
            "height": cls._parse_height_weight(line_height, key="height", unit="cm"),
            "weight": cls._parse_height_weight(line_weight, key="weight", unit="kg"),
            "values": cls._parse_values(lines_value),
            "comments": cls._parse_comments(lines_comment),
            **cls._parse_patient_metadata(line_pmd),
        }
        lx_metadata: LXMetaData = LXMetaData(**md_dict)
        return lx_metadata

    @staticmethod
    def _parse_patient_metadata(
        line: Optional[str], separator: str = ","
    ) -> Dict[str, Union[str, bool]]:
        """Parse patient metadata from metadata line.

        # Nahrungskarenz: über 3 Std., Raucher: Nein, Sauerstoff: Nein, Beatmung: Nein, Medikation: Ja
        """
        if not line:
            return {
                "food_abstinence": "-",
                "smoking": False,
                "oxygen": False,
                "ventilation": False,
                "medication": False,
            }
        else:
            tokens = line.split(separator)
            d: Dict[str, str] = {}
            for k, key in enumerate(MetadataParser.patient_metadata_fields):
                try:
                    d[key] = tokens[k].split(":")[1].strip()
                except IndexError:
                    logger.error(f"'{key}' could not be parsed from '{tokens[k]}'")
                    d[key] = "-"

            d_processed: Dict[str, Union[str, bool]] = {**d}
            for key in ["smoking", "oxygen", "ventilation", "medication"]:
                if d[key].lower() == "ja":
                    d_processed[key] = True
                elif d[key].lower() == "nein":
                    d_processed[key] = False
                else:
                    logger.error(f"Invalid value in metadata: '{key}: {d[key]}'")
                    d_processed[key] = False

            if d["food_abstinence"] == "über 3 Std.":
                d_processed["food_abstinence"] = "> 3 hr"
            return d_processed

    @staticmethod
    def _parse_mid(line: Optional[str]) -> str:
        """Parse mid."""
        mid: str = "-"
        if not line:
            logger.error("No 'mid' information in LiMAx metadata")
            return mid

        if line.startswith("mID "):
            mid = line[4:]
        else:
            mid = line

        return mid

    @staticmethod
    def _parse_datetime(line: Optional[str]) -> str:
        """Parse datetime information.

        Example: 01.01.2010 08:30
        """
        if not line:
            logger.warning("No 'datetime' information in LiMAx metadata")
            return "-"

        date_str = line
        try:
            # Check that date can be parsed
            date = datetime.strptime(line, "%d.%m.%Y %H:%M")
            date_str = date.strftime("%Y-%m-%d %H:%M")

        except ValueError:
            logger.error(
                f"'datetime' could not be parsed in format '%d.%m.%Y %H:%M' from '{line}"
            )

        return date_str

    @staticmethod
    def _parse_sex(line: Optional[str]) -> str:
        """Parse sex in {M, F, NA}."""
        if not line:
            # no sex information in the file
            return "NA"
        if line == "männlich":
            return "M"
        elif line == "weiblich":
            return "F"
        else:
            logger.error(f"'sex' could not be parsed from '{line}'")
            return "NA"

    @staticmethod
    def _parse_height_weight(line: Optional[str], key: str, unit: str) -> float:
        """Parse patient height or weight."""
        value: float = -1.0
        if not line:
            logger.warning(f"No {key} information in LiMAx metadata")
            return value

        try:
            tokens = line.split()
            value = float(tokens[0])
            if tokens[1] != unit:
                logger.error(f"'{key}' unit is not '{unit}', but '{tokens[1]}'")
        except (IndexError, ValueError) as err:
            logger.error(f"'{key}' could not be parsed from '{line}'. {err}")
        return value

    @staticmethod
    def _parse_values(lines: List[str]) -> List[float]:
        """Parse value fields."""
        values: List[float] = []
        for line in lines:
            try:
                line = line.replace(",", ".")
                values.append(float(line))
            except ValueError as err:
                logger.warning(f"'value' could not be parsed from '{line}'. {err}")
        return values

    @staticmethod
    def _parse_comments(lines: List[str]) -> List[str]:
        """Parse comments."""
        return lines


def parse_limax_file(
    limax_csv: Path,
    output_dir: Optional[Path] = None,
) -> LX:
    """Read limax data."""
    if output_dir:
        output_dir.mkdir(parents=True, exist_ok=True)
        # limax_path = output_dir / f"{limax_csv.stem}.txt"
        json_path = output_dir / f"{limax_csv.stem}.json"
        fig_path = output_dir / f"{limax_csv.stem}.png"
        console.rule(
            title=f"[bold white]Process '{limax_csv}'", align="left", style="[white]"
        )

    # parse file
    line_offset = -1
    lines: List[str] = []

    with open(limax_csv, "r") as f:
        raw_lines: List[str] = f.readlines()

        # cleanup lines
        for line in raw_lines:
            # remove comment character
            if line.startswith("# "):
                line = line[2:]
            # strip whitespace
            line = line.strip()
            # remove empty lines
            if len(line) > 0:
                lines.append(line)

        # find metadata offset in clean lines
        for k, line in enumerate(lines):
            if line.startswith("Zeit"):
                line_offset = k
                break
        else:
            raise ValueError(
                f"No line starting with 'Zeit' in csv, invalid LIMAX file: {limax_csv}"
            )

    md_lines = lines[:line_offset]
    data_lines = lines[line_offset + 1 :]

    # parse metadata
    lx_metadata: LXMetaData = MetadataParser.parse(md_lines)

    # parse data
    time, dob, error = [], [], []
    for line in data_lines:
        # cleanup of lines
        tokens = [t.strip() for t in line.split("\t")]
        time.append(int(tokens[0]))
        dob.append(float(tokens[1]))
        error.append(str(tokens[2]))

    d: Dict[str, Any] = {
        "time": time,
        "dob": dob,
        "error": error,
    }
    df = pd.DataFrame(data=d)
    df = df[["time", "dob", "error"]]

    # sort by time (some strange artefacts in some files)
    df.sort_values(by=["time"], inplace=True)
    lx_data = LXData(
        time=list(df.time.values), dob=list(df.dob.values), error=list(df.error.values)
    )
    lx = LX(metadata=lx_metadata, data=lx_data)
    console.print(lx_metadata)

    # serialization to JSON
    if output_dir:
        with open(json_path, "w") as f_json:
            json_str = lx.model_dump_json(indent=2)
            f_json.write(json_str)
        plot_lx_matplotlib(lx, fig_path=fig_path)
        console.print(f"Json: file://{json_path}")
        console.print(f"Figure: file://{fig_path}")

    return lx


if __name__ == "__main__":
    from limax import (
        EXAMPLE_LIMAX_PATIENT1_PATH,
        EXAMPLE_LIMAX_PATIENT2_PATH,
        EXAMPLE_LIMAX_PATIENT3_PATH,
        EXAMPLE_LIMAX_PATIENT2017_PATH,
        EXAMPLE_LIMAX_PATIENT2018_PATH,
        EXAMPLE_LIMAX_PATIENT2022_PATH,
        PROCESSED_DIR,
        RAW_DIR,
    )

    for path in [
        EXAMPLE_LIMAX_PATIENT1_PATH,
        EXAMPLE_LIMAX_PATIENT2_PATH,
        EXAMPLE_LIMAX_PATIENT3_PATH,
        EXAMPLE_LIMAX_PATIENT2017_PATH,
        EXAMPLE_LIMAX_PATIENT2018_PATH,
        EXAMPLE_LIMAX_PATIENT2022_PATH,
    ]:
        lx = parse_limax_file(limax_csv=path, output_dir=PROCESSED_DIR)
        # console.print(lx)
        # console.print(lx.json())
        console.print()
        console.print(lx.data.to_df().head(5))

    read_limax_dir(input_dir=RAW_DIR, output_dir=PROCESSED_DIR)
